; PID control for two wheel robot on SX (Ubicom) microcontroller.
; Copyright (C) 1999-2001 Blaise Gassend <blaise at gassend dot com>
; 
; This program is free software; you can redistribute it and/or
; modify it under the terms of the GNU General Public License
; as published by the Free Software Foundation; either version 2
; of the License, or (at your option) any later version.
; 
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
; 
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

; cd ~/robot/sources/picasm/;picasm servo.asm  

	device pic16c73a
	config  CP=off,WDT=off,PWRT=off,OSC=hs

include "myinclude.h"
include "mymacro.h"

_SEL equ 1
_OE equ 3
_DIR equ 2


; Bank1
temp 	equ 	20h ; 20h -> 23h
temp1 	equ 	22h ; Redundant with temp
temp2 	equ 	23h ; Redundant with temp
recpos 	equ 	24h

; LSB first!
targpos equ 	26h ; 25h -> 28h (24:8 bits)
pos	equ 	29h ; 29h -> 2Bh (24:0 bits)
ierr	equ 	2Ch ; 2Ch -> 2Fh (32:0 bits) ; Only 3 high bytes added in filter.
perr    equ 	30h ; 30h -> 32h (24:0 bits)

speed   equ 	37h ; 34h -> 38h (16:24 bits)
speedhi equ 	38h ; Redundant with speed
speedlo equ 	37h ; Redundant with speed
pcount  equ	3Ah ; when to add p
icount  equ	3Bh ; when to add i
dcount  equ	3Ch ; when to add d
frict	equ	3Dh ; 3Dh -> 3Eh (16:0 bits)
recbuf	equ	3Fh ; 3Fh -> 42h (32 bits)
targspd equ	44h ; 43h -> 45h (16:8 bits)
reqspd 	equ	46h ; 46h -> 47h (16:0 bits)
reqacc	equ	49h ; 48h -> 4Ah (16:8 bits)
derr	equ 	4Bh ; 4Bh -> 4Ch (8:0 bits)
ierrlim equ	4Dh ; 4Dh -> 4Eh (16:16:0 bits)

outcnt	equ	70h

	org 	0
	bcf 	INTCON,GIE

	bsf	STATUS,RP0 ; set PR2 (timer count from) to 0FFh
        movlw	0FFh
	movwf	PR2

	bcf     STATUS,RP0 ; set duty cycle
	clrf	speed
	clrf	speed+1
	clrf	targpos-1
	clrf	targpos
	clrf	targpos+2
	movlw	80h
	movwf	targpos+1
	clrf	targspd-1
	clrf	targspd
	clrf	targspd+1
	clrf	ierr
	clrf	ierr+1
	clrf	ierr+2
	clrf	ierr+3
	clrf	frict
	clrf	pos+2
	movlf	ierrlim,44
	movlf	ierrlim+1,44
	movlf	frict+1,1
	movlw	24
	movwf	pcount
	movlw   23
	movwf	icount
	movlw   28
	movwf	dcount
	clrf	outcnt
	clrf	recpos
	call	SetSpeed

	bsf	STATUS,RP0 ; Output on PWM port and serial TX
	bcf	TRISC,2
	bcf	TRISC,6

	bcf     STATUS,RP0 ; Set clock mode
        bcf	T2CON,T2CKPS1
	bcf	T2CON,T2CKPS0
	bsf	T2CON,TMR2ON

	bsf	CCP1CON,CCP1M3 ; Set to PWM mode
	bsf	CCP1CON,CCP1M2

        bsf     STATUS,RP0 ; Not analog to digital mode, out on A2 and A3
	movlw	7
	movwf	ADCON1
        bcf	TRISA,_SEL
        bcf	TRISA,_OE
        bcf	TRISA,_DIR

        movlw	7
	
	movlw	7 ; serial
	movwf	SPBRG
	bcf	TXSTA,SYNC
	bcf	STATUS,RP0
	bsf	RCSTA,SPEN
	bsf	RCSTA,CREN
	bsf	STATUS,RP0
        bsf     TXSTA,TXEN
	bcf	STATUS,RP0

Loop
        bcf	RCSTA,OERR ; Clear any serial errors

	call	WaitSerial

	movf	outcnt,F
	btfsc	STATUS,Z
	movlf	outcnt,100
	decf	outcnt,F

	clrf	TXREG

	call	WaitSerial
	
	movf	CCPR1L,0
	movwf	TXREG

	call	WaitSerial
		
	clrw	
	btfsc	PORTA,_DIR
	movlw	80h	
	iorwf	CCP1CON,0
	movwf	TXREG

	call	WaitSerial

;	bcf	PORTA,_SEL ; read high byte of position
;	bcf	PORTA,_OE
;	nop
;       movf	PORTB,0
;	movwf	temp+1

;	bsf	PORTA,_SEL ; read low byte of position
;	nop
;	movf	PORTB,0
;	movwf	temp
;	bsf	PORTA,_OE

	bsf	PORTA,_OE ; Latch counter
	nop
	bcf	PORTA,_OE ; Unlatch counter
	nop
	bcf	PORTA,_SEL ; read first counter
	nop
	movff	temp1,PORTB
	bsf	PORTA,_SEL ; read second counter
	nop
	movff	temp2,PORTB

	addff	pos,temp1 ; Add first counter movement
	clrw
	btfsc	temp1,7
	movlw	0ffh
	addcwf	pos+1
	clrw
	btfsc	temp1,7
	movlw	0ffh
	addcwf	pos+2
	
	addff	targpos,temp2 ; Add second counter movement
	clrw
	btfsc	temp2,7
	movlw	0ffh
	addcwf	targpos+1
	clrw
	btfsc	temp2,7
	movlw	0ffh
	addcwf	targpos+2

;	subff	temp,pos ; Add delta pos to pos
;	subcff	temp+1,pos+1
;	addff	pos,temp
;	addcff	pos+1,temp+1
;	clrw
;	btfsc	temp+1,7
;	movlw	0ffh
;	addcwf	pos+2

;	addff	targpos-1,targspeed-1; Add targspd to targpos
;	addcff	targpos,targspeed
;	addcff	targpos+1,targspeed
;	clrw
;	btfsc	targpos+1,7
;	movlw	0ffh
;	addcwf	targpos+1

	call	PIDFilter
	call	SetSpeed

	movff	TXREG,speed+1
	call	WaitSerial

	movff	TXREG,speed 
	call	WaitSerial

	movff	TXREG,pos+2
	call	WaitSerial

	movff	TXREG,pos+1
	call	WaitSerial

	movff	TXREG,pos
	call	WaitSerial

;	movff	TXREG,ierr+3
;	call	WaitSerial

;	movff	TXREG,ierr+2
;	call	WaitSerial

;	movff	TXREG,ierr+1
;	call	WaitSerial

	movff	TXREG,perr+2
	call	WaitSerial

	movff	TXREG,perr+1
	call	WaitSerial

	movff	TXREG,perr
	call	WaitSerial

;	movff	TXREG,derr
;	call	WaitSerial

;	movff	TXREG,derr+1
;	call	WaitSerial

	goto	Loop
		
	
WaitSerial ; Waits for output to be sent and recieve any commands
	bcf	STATUS,RP0
        btfss	PIR1,RCIF
	goto	NoReceive

	movlw	recbuf
	addwf	recpos,W
	movwf	FSR
	movff	INDF,RCREG
	incf	recpos,F
	movlw	4
	subwf	recpos,W
	btfss	STATUS,Z
	goto	NoReceive
	clrf	recpos
	movff	targpos,recbuf
	movff	targpos+1,recbuf+1
	movff	targpos+2,recbuf+2
	movff	targpos+3,recbuf+3

NoReceive
        btfss	PIR1,TXIF
        goto    WaitSerial
	return

SetSpeed ; Sets the PWM speed from speedhi and speedlo
	bcf	STATUS,RP0	
	movf	speedhi,0
	movwf	temp2
	movf	speedlo,0
	movwf	temp1
	btfss	speedhi,7
	goto	PosDirection
	comf	temp2,1
	comf	temp1,1
	incf	temp1,1
	btfsc	STATUS,Z
	incf	temp2,1
	bsf	PORTA,_DIR
	goto	SetSpeedTestSize
PosDirection
	bcf	PORTA,_DIR
SetSpeedTestSize
	movf	temp1,F
	btfsc	STATUS,Z
	movf	temp2,F
	btfsc	STATUS,Z
	goto	IsZero
	addff	temp1,frict
	addcff	temp2,frict+1
IsZero:
	movlw	0ffh
	btfsc	STATUS,C
	movwf	temp2
	movlw	0fch
	andwf	temp2,0
	btfsc	STATUS,Z
	goto	SetSpeedAct
	movlw	0ffh
	movwf	temp1
	movlw	03h
	movwf	temp2
SetSpeedAct
	bcf	CCP1CON,CCP1X
	btfsc	temp1,1
	bsf	CCP1CON,CCP1X
	bcf	CCP1CON,CCP1Y
	btfsc	temp1,0
	bsf	CCP1CON,CCP1Y
	rrf	temp2,1
	rrf	temp1,1
	rrf	temp2,1
	rrf	temp1,0
	movwf	CCPR1L
	return
	
PIDFilter
	; Calculate derr = -perr
	negff	derr,perr
	negcff	derr+1,perr+1
	; Calculate perr = targpos - pos
	movff	perr+1,targpos+1
	movff	perr+2,targpos+2
	subfff  perr,targpos,pos
	subcff  perr+1,pos+1
	sublcff perr+2,pos+2
	; Calculate ierr = ierr + perr
	addff 	ierr,perr	
	addcff 	ierr+1,perr+1
	addcff 	ierr+2,perr+2
	btfsc 	perr+2,7
	decf 	ierr+3
	btfsc 	STATUS,C
	incf 	ierr+3
	; Saturate ierr with ierrlim
	btfsc	ierr+3,7
	goto	IErrSatNeg
	movf	ierrlim+1,W
	subwf	ierr+3,W
	btfss	STATUS,C
	goto	IErrSatDone
	btfss	STATUS,Z
	goto	IErrSatHigh
	movf	ierrlim,W
	subwf	ierr+2,W
	btfss	STATUS,C
	goto 	IErrSatDone
	btfsc	STATUS,Z
	goto	IErrSatDone
IErrSatHigh
	movlw	0ffh
	movwf	ierr
	movwf	ierr+1
	movff	ierr+2,ierrlim
	movff	ierr+3,ierrlim+1
	goto	IErrSatDone
IErrSatNeg
	movf	ierrlim+1,W
	addwf	ierr+3,W
	btfsc	STATUS,C
	goto	IErrSatDone
	addlw	1
	btfss	STATUS,Z
	goto	IErrSatLow
	movf	ierrlim,W
	addwf	ierr+2,W
	btfsc	STATUS,C
	goto	IErrSatDone
	addlw	1
	btfsc	STATUS,Z
	goto	IErrSatDone
IErrSatLow
	clrf	ierr
	clrf	ierr+1
	comf	ierrlim,W
	movwf	ierr+2
	comf	ierrlim+1,W
	movwf	ierr+3
IErrSatDone	
	; Calculate derr = derr + perr
	addff 	derr,perr
	addcff	derr+1,perr+1
	; Calculate speed = 0;
	clrf	speed-3
	clrf	speed-2
	clrf	speed-1
	clrf	speed
	clrf	speed+1
	; Do the summing
	movlw	34
	movwf	temp
	clrf	temp+1
SumPIDLoop
	cmpff	temp,pcount
	btfsc	STATUS,Z
	call	PIDAddP
	cmpff	temp,icount
	btfsc	STATUS,Z
	call	PIDAddI
	cmpff	temp,dcount
	btfsc	STATUS,Z
	call	PIDAddD
	rlf	speed+1,W ; Detect overflow
	xorwf	speed+1,W
	andlw	80h
	btfss	STATUS,Z
	goto	PIDSpeedOverflow	
	bcf	STATUS,C
	rlf	speed-3,F
	rlf	speed-2,F
	rlf	speed-1,F
	rlf	speed,F
	rlf	speed+1,F
	decfsz	temp
	goto 	SumPIDLoop
	return	
PIDSpeedOverflow
	clrw
	btfss	speed+1,7
	movlw	0FFh
	movwf	speed
	xorlw	80h
	movwf	speed+1
	return
	
PIDAddD
	addff	speed-3,derr
	addcff	speed-2,derr+1
	btfsc	derr+1,7
	goto	PIDSubCarryFrom1
	btfss	STATUS,C
	return
	incfsz	speed-1,F
	return
	incfsz	speed,F
	return
	incf	speed+1
	movlw	80h
	subwf	speed+1,W
	btfsc	STATUS,Z
	decf	speed+1
	return
PIDSubCarryFrom1
	btfsc	STATUS,C
	return  
	movlw	1
	subwf	speed-1,F
	btfsc	STATUS,C
	return
	subwf	speed,F
	btfsc	STATUS,C
	return
	subwf	speed+1,F
	movlw	7Fh
	subwf	speed+1,W
	btfss	STATUS,Z
	return
	incf	speed+1
	return

PIDAdd	macro
	addff	speed-3,\1
	addcff	speed-2,\1+1
	addcff	speed-1,\1+2
	btfsc	\1+2,7
	goto	PIDSubCarryFrom0
	goto	PIDAddCarryFrom0
	endm

PIDAddI
	PIDAdd	ierr+1

PIDAddP 
	PIDAdd	perr

PIDAddCarryFrom0
	btfss	STATUS,C
	return
	incfsz	speed,F
	return
	incf	speed+1
	movlw	80h
	subwf	speed+1,W
	btfsc	STATUS,Z
	decf	speed+1
	return
PIDSubCarryFrom0
	btfsc	STATUS,C
	return  
	movlw	1
	subwf	speed,F
	btfsc	STATUS,C
	return
	subwf	speed+1,F
	movlw	7Fh
	subwf	speed+1,W
	btfss	STATUS,Z
	return
	incf	speed+1
	return
