; PID control for two wheel robot on SX (Ubicom) microcontroller.
; Copyright (C) 1999-2001 Blaise Gassend <blaise at gassend dot com>
; 
; This program is free software; you can redistribute it and/or
; modify it under the terms of the GNU General Public License
; as published by the Free Software Foundation; either version 2
; of the License, or (at your option) any later version.
; 
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
; 
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

; Derived constants
IERRLIMH = (IERRLIM >> 8)
IERRLIML = (IERRLIM & 255)D
NIERRLIMH = ((-IERRLIM-1) >> 8)
NIERRLIML = ((-IERRLIM-1) & 255)

;;;     include "macro.inc"

	org     200h

CallPID
	bank    INT_B
	; copy dists to pids and adjust counts to 0080h
getspeed1
	mov     temp+1,count1+1
	mov     temp,count1 ; count1 to temp
	stc
	mov     W,temp+1
	mov     W,count1+1-W
	jnz     getspeed1
	stc
	sub     count1,temp
	sub     count1+1,temp+1
getspeed2
	mov     temp+2,count2 ; count2 to temp
	mov     temp+3,count2+1
	stc
	mov     W,temp+3
	mov     W,count2+1-W
	jnz     getspeed2
	stc
	sub     count2,temp+2
	sub     count2+1,temp+3
	bank    PID1_B ; temp to pid1
	clc
	mov     W,temp
	mov     dist,W
	add     abspos,W
	mov     W,temp+1
	mov     dist+1,W
	add     abspos+1,W
	clr     W
	snb     temp+1.7
	mov     W,#0FFh
	add     abspos+2,W
	bank    PID2_B ; temp to pid2
	clc
	mov     W,temp+2
	mov     dist,W
	add     abspos,W
	mov     W,temp+3
	mov     dist+1,W
	add     abspos+1,W
	clr     W
	snb     temp+3.7
	mov     W,#0FFh
	add     abspos+2,W
	page    PIDFilter
	call    PIDFilter ; execute pid2
	bank    PID1_B
	page    PIDFilter
	call    PIDFilter ; execute pid1
	mov     temp,ispeed ; copy ispeed1 to temp
	mov     temp+1,ispeed+1
	bank    PID2_B 
	mov     temp+2,ispeed ; copy ispeed2 to temp
	mov     temp+3,ispeed+1

	bank    INT_B
	jnb     temp+1.7,noreverse1 ; Update PWM1
	not     temp    ; Invert value in temp and change direction
	not     temp+1
	clr     W
	stc
	add     temp,W
	add     temp+1,W
	setb    ra.DIR1
	jmp     friction1
noreverse1
	clrb    ra.DIR1
friction1
	mov     W,temp
	or      W,temp+1
	jz      setpwm1
	clc ; Compensate friction and saturate
	add     temp,#(FRICTION & 255)
	add     temp+1,#(FRICTION >> 8)
	mov     W,#0FCh 
	and     W,temp+1
	jz      setpwm1
	mov     temp+1,#3h
	mov     temp,#0FFh
setpwm1
	mov     speed1,temp
	mov     speed1+1,temp+1

	jnb     temp+3.7,noreverse2 ; Update PWM2
	not     temp+2  ; Invert value in temp and change direction
	not     temp+3
	clr     W
	stc
	add     temp+2,W
	add     temp+3,W
	setb    ra.DIR2
	jmp     friction2
noreverse2
	clrb    ra.DIR2
friction2
	mov     W,temp+2
	or      W,temp+3
	jz      setpwm2
	clc ; Compensate friction and saturate
	add     temp+2,#(FRICTION & 255)
	add     temp+3,#(FRICTION >> 8)
	mov     W,#0FCh 
	and     W,temp+3
	jz      setpwm2
	mov     temp+3,#3h
	mov     temp+2,#0FFh
setpwm2
	mov     speed2,temp+2
	mov     speed2+1,temp+3

	ret

;******************************************************
	
addp
	addnc   fspeed,  perr
	addc    fspeed+1,perr+1
	addc    fspeed+2,perr+2
;sextow perr+2
 clr    w ; bug in sasm
 snb    perr+2.7
 mov    W,#0FFh
	addWc   fspeed+3
	addWc   fspeed+4
	ret     

addi
	addnc   fspeed,  ierr+1
	addc    fspeed+1,ierr+2
	addc    fspeed+2,ierr+3
;sextow ierr+3
 clr    w ; bug in sasm
 snb    ierr+3.7
 mov    W,#0FFh
	addWc   fspeed+3
	addWc   fspeed+4
	ret     

addd
	addnc   fspeed,  derr
	addc    fspeed+1,derr+1
;sextow derr+1
 clr    w ; bug in sasm
 snb    derr+1.7
 mov    W,#0FFh
	addWc   fspeed+2
	addWc   fspeed+3
	addWc   fspeed+4
	ret     

shifttotal ; shift and check for possible overflows
	clc
	rl      fspeed
	rl      fspeed+1
	rl      fspeed+2
	rl      fspeed+3
	rl      fspeed+4
	mov     W,<<fspeed+4
	xor     W,fspeed+4
	and     W,#0C0h 
	clc
	snz
	ret     ; Three first bytes equal so no undetectable overflows
	stc
	ret
	
;***************************************

PIDFilter
	; Remove 1 from ierr
	mov     W,perr
	or      W,perr+1
	or      W,perr+2
	jnz     NoTamperIErr
	mov     W,ierr
	or      W,ierr+1
	or      W,ierr+2
	or      W,ierr+3
	jz      NoTamperIErr
	clr     W
	clc      
	sb      ierr+3.7
	mov     W,#0FFh
	snb     ierr+3.7
	stc
	add     ierr,W
	add     ierr+1,W
	add     ierr+2,W
	add     ierr+3,W
	stc
	sub     ierr,ierr+1
	sub     ierr+1,ierr+2
	sub     ierr+2,ierr+3
	clr     W
	snb     ierr+3.7
	mov     W,#0FFh
	sub     ierr+3,W
NoTamperIerr
	; Calculate derr
	; derr = derr >> 1
	mov     W,<<derr+1 ; carry = derr.15
	rr      derr+1
	rr      derr
	clrb    Z
	snb     derr+1.7
	inc     derr
	snz
	inc     derr+1
	addnc   derr,  dist ; derr = derr + dist
	addc    derr+1,dist+1

	; Calculate perr
	addnc   perr,  dist ; perr = perr + dist
	addc    perr+1,dist+1
;sextow         dist+1
 clr    w ; bug in sasm
 snb    dist+1.7
 mov    W,#0FFh
	addWc   perr+2

	; Calculate ierr
	clc
	add   ierr,  perr ; ierr = ierr + perr
	add    ierr+1,perr+1
	add    ierr+2,perr+2
;sextow         perr+2
 clr    w ; bug in sasm
 snb    perr+2.7
 mov    W,#0FFh
	add   ierr+3,w

	; ierr = saturate(ierr)
	jb      ierr+3.7,ierrSatLow
	cmp     ierr+3,IERRLIMH
	jnc     ierrSatEnd ; jump below
	jnz     ierrTooHigh ; jump above
	cmp     ierr+2,IERRLIML
;;;;    jabove  ierrTooHigh
	jnc     ierrSatEnd ; jump below
	jnz     ierrTooHigh ; jump above
	jmp     ierrSatEnd
ierrSatLow
	cmp     ierr+3,NIERRLIMH
	jnc     ierrTooLow ; jump below
	jnz     ierrSatEnd ; jump above 
	cmp     ierr+2,NIERRLIML
	jnc     ierrTooLow ; jump below
	jmp     ierrSatEnd
ierrTooHigh
	mov     W,#0FFh
	mov     ierr,W
	mov     ierr+1,W
	mov     W,#IERRLIML
	mov     ierr+2,W
	mov     W,#IERRLIMH
	mov     ierr+3,W
	jmp     ierrSatEnd
ierrTooLow
	clr     ierr
	clr     ierr+1
	mov     W,#NIERRLIML
	mov     ierr+2,W
	mov     W,#NIERRLIMH
	mov     ierr+3,W
ierrSatEnd      
	; speed = 0
	clr     ispeed-4
	clr     ispeed-3
	clr     ispeed-2
	clr     ispeed-1
	clr     ispeed
	clr     ispeed+1
	; Do the adding...

	; bit 32
IF (PCONSTH)
	call    addp
ENDIF
IF (ICONSTH)
	call    addi
ENDIF
IF (DCONSTH)
	call    addd
ENDIF
IF ((DCONSTH || PCONSTH) || ICONSTH)
	call    shifttotal
	snc
	ret     
ENDIF

	; bits 31:16
step = 15
REPT    16
IF ((PCONST >> step) & 1)
	call    addp
ENDIF
IF ((ICONST >> step) & 1)
	call    addi
ENDIF
IF ((DCONST >> step) & 1)
	call    addd
ENDIF
IF ((((((DCONST >> step) || (PCONST >> step)) || (ICONST >> step)) || PCONSTH) || ICONSTH) || DCONSTH)
	call    shifttotal
	snc
	ret     
ENDIF
step = step - 1
ENDR

	; bits 15:0
step = 15
REPT    16
IF ((PCONSTL >> step) & 1)
	call    addp
ENDIF
IF ((ICONSTL >> step) & 1)
	call    addi
ENDIF
IF ((DCONSTL >> step) & 1)
	call    addd
ENDIF
IF (((((((((DCONSTL >> step) || (PCONSTL >> step)) || (ICONSTL >> step)) || PCONST) || ICONST) || DCONST) || PCONSTH) || ICONSTH) || DCONSTH)
	call    shifttotal
	snc
	ret     
ENDIF
step = step - 1
ENDR

	; We've added!
	ret

pagepidend
	

