; PID control for two wheel robot on SX (Ubicom) microcontroller.
; Copyright (C) 1999-2001 Blaise Gassend <blaise at gassend dot com>
; 
; This program is free software; you can redistribute it and/or
; modify it under the terms of the GNU General Public License
; as published by the Free Software Foundation; either version 2
; of the License, or (at your option) any later version.
; 
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
; 
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

; Constants
IERRLIM = 4444h
PCONSTH = 0
PCONST = 1
PCONSTL = 0
ICONSTH = 0
ICONST = 1
ICONSTL = 0
DCONSTH = 0
DCONST = 1 
DCONSTL = 0

; Derived constants
IERRLIMH = (IERRLIM >> 8)
IERRLIML = (IERRLIM & 255)
NIERRLIMH = ((-IERRLIM-1) >> 8)
NIERRLIML = ((-IERRLIM-1) & 255)

org	9h ; locals

org	0dh
fspeed	ds	3 ; fracional part of speed
org	10h ; globals
ispeed  ds	2 ; integer part of speed
perr	ds	3
ierr	ds	4
derr	ds	2
dist	ds	2 ; moved since last time

	include "macro.inc"

	org	200h
PIDFilter
	; Calculate derr
	movb	C,derr.7 ; derr = derr >> 1
	rr	derr+1
	rr	derr
	addnc	derr,  dist ; derr = derr + dist
	addc	derr+1,dist+1

	; Calculate perr
	addnc	perr,  derr ; perr = perr + dist
	addc	perr+1,derr+1
	sextow 	derr+1
	addWc	perr+2

	; Calculate ierr
	addnc	ierr,  perr ; ierr = ierr + perr
	addc	ierr+1,perr+1
	addc	ierr+2,perr+2
	sextow 	perr+2
	addWc	ierr+3

	; ierr = saturate(ierr)
	jb	ierr+3.7,ierrSatLow
	cmp	ierr+3,IERRLIMH
	jabove	ierrTooHigh
	cmp	ierr+2,IERRLIML
	jabove	ierrTooHigh
	jmp	ierrSatEnd
ierrSatLow
	cmp	ierr+3,NIERRLIMH
	jbelow	ierrTooLow
	cmp	ierr+2,NIERRLIML
	jbelow	ierrTooLow
	jmp	ierrSatEnd
ierrTooHigh
	mov	W,#0FFh
	mov	ierr,W
	mov	ierr+1,W
	mov	W,#IERRLIML
	mov	ierr+2,W
	mov	W,#IERRLIMH
	mov	ierr+3,W
	jmp	ierrSatEnd
ierrTooLow
	clr	ierr
	clr	ierr+1
	mov	W,#NIERRLIML
	mov	ierr+2,W
	mov	W,#NIERRLIMH
	mov	ierr+3,W
ierrSatEnd	
	; speed = 0
	clr	ispeed-4
	clr	ispeed-3
	clr	ispeed-2
	clr	ispeed-1
	clr	ispeed
	clr	ispeed+1
	; Do the adding...

	; bit 32
IF (PCONSTH)
	call	addp
ENDIF
IF (ICONSTH)
	call	addi
ENDIF
IF (DCONSTH)
	call	addd
ENDIF
IF ((DCONSTH || PCONSTH) || ICONSTH)
	call	shifttotal
ENDIF

	; bits 31:16
step = 15
REPT	16
IF ((PCONST >> step) & 1)
	call	addp
ENDIF
IF ((ICONST >> step) & 1)
	call	addi
ENDIF
IF ((DCONST >> step) & 1)
	call	addd
ENDIF
IF ((((((DCONST >> step) || (PCONST >> step)) || (ICONST >> step)) || PCONSTH) || ICONSTH) || DCONSTH)
	call	shifttotal
ENDIF
step = step - 1
ENDR

	; bits 15:0
step = 15
REPT	16
IF ((PCONSTL >> step) & 1)
	call	addp
ENDIF
IF ((ICONSTL >> step) & 1)
	call	addi
ENDIF
IF ((DCONSTL >> step) & 1)
	call	addd
ENDIF
IF (((((((((DCONSTL >> step) || (PCONSTL >> step)) || (ICONSTL >> step)) || PCONST) || ICONST) || DCONST) || PCONSTH) || ICONSTH) || DCONSTH)
	call	shifttotal
	snc
	ret	
ENDIF
step = step - 1
ENDR

	; We've added!
	ret

; ##############################################	

addp
	addnc	fspeed,  ierr
	addc	fspeed+1,ierr+1
	addc	fspeed+2,ierr+2
	sextow	derr+2
	addWc	fspeed+3
	addWc	fspeed+4
	ret	

addi
	addnc	fspeed,  derr
	addc	fspeed+1,derr+1
	addc	fspeed+2,derr+2
	addc	fspeed+3,derr+3
	sextow	derr+3
	addWc	fspeed+4
	ret	

addd
	addnc	fspeed,  derr
	addc	fspeed+1,derr+1
	sextow	derr+1
	addWc	fspeed+2
	addWc	fspeed+3
	addWc	fspeed+4
	ret	

shifttotal
	clc
	rl	fspeed
	rl	fspeed+1
	rl	fspeed+2
	rl	fspeed+3
	rl	fspeed+4
	mov	W,<<fspeed+4
	xor	W,fspeed+4
	xor	W,#0C0h 
	clc
	snz
	ret	; Three first bytes equal so no undetectable overflows
	stc
	ret
	
