; PID control for two wheel robot on SX (Ubicom) microcontroller.
; Copyright (C) 1999-2001 Blaise Gassend <blaise at gassend dot com>
; 
; This program is free software; you can redistribute it and/or
; modify it under the terms of the GNU General Public License
; as published by the Free Software Foundation; either version 2
; of the License, or (at your option) any later version.
; 
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
; 
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

; Various flags

Ch1 	equ	0
Ch2	equ	1
ResetPID equ	2
ResetTrap equ	3
;;; NoStopBit equ	4

org	600h

ExecSerCmd
	mov	M,#7 ; For range functions
	mov	FSR,#30h
	mov	W,#0D0h ; Test if recieved cmd or args.
	and	W,iostat
	xor	W,#0D0h
	sz	
	ret
	jb	iostat.5,RecievedArgs
	mov	W,iobyte	; We have just recieved a command
	mov	cmd,W
	jz	cmd00
	xor	W,#(01h ^ 00h)
	jz	cmd01
	xor	W,#(10h ^ 01h)
	jz	cmd10
	xor	W,#(20h ^ 10h)
	jz	cmd20
	xor	W,#(0FDh ^ 20h)
	jz	cmdFD
	xor	W,#(0FEh ^ 0FDh)
	jz	cmdFE
	xor	W,#(0FFh ^ 0FEh)
	jz	cmdFF
	xor	W,#(011h ^ 0FFh)
	jz	cmd11
	xor	W,#(012h ^ 011h)
	jz	cmd12
	jmp	nextcommand

RecievedArgs
	mov	W,cmd
	xor	W,#01h
	jz	arg01
	xor	W,#(10h ^ 01h)
	jz	arg10
	xor	W,#(20h ^ 10h)
	jz	arg20
	xor	W,#(011h ^ 020h)
	jz	arg11
	xor	W,#(012h ^ 011h)
	jz	arg12

cmd00
nextcommand
	bank	#SPI_B
	and 	iostat,#0Fh
	or	iostat,#0E0h
	mov	spidelay,#0FFh
	mov	iobyte,#WAITINGFORCMD
	jmp	startupspi

getargs
	bank	#SPI_B
	mov	temp,W
	and 	iostat,#0Fh
	mov	W,<>temp
	or	iostat,W
	mov	spidelay,#0FFh
	mov	FSR,#30h
	add	FSR,temp
	mov	iobyte,INDF

startupspi
	mov	W,#0Fh
	and	W,iostat
	snz
	ret
	page	StartSPI
	call	StartSPI
	ret

PrepRange
	; M:W -> prog address to read
	; read M -> count, read W-> address
	; restores M, W, FSR at end

	; temp -> W
	; temp+1 -> M
	; temp+2 -> FSR
	; temp+3 -> Dest
	; temp+4 -> Count

	mov	temp,W
	mov	W,M
	mov	temp+1,W
	mov	temp+2,FSR
	mov	W,temp
	iread
	mov	temp+3,W
	mov	W,M
	mov	temp+4,W
	ret

EndRange
	mov	FSR,temp+2
	mov	W,temp+1
	mov	M,W
	mov	W,temp
	ret

ClrRange
	call	PrepRange
	mov	FSR,temp+3
ClrRangeLoop
	clr	INDF
	inc	FSR
	decsz	temp+4
	jmp	ClrRangeLoop
	call	EndRange
	ret

AddRange
	call	PrepRange
	clc
AddRangeLoop
	mov	FSR,temp+2
	mov	temp+5,INDF
	mov	FSR,temp+3
	add	INDF,temp+5
	inc	temp+2
	inc	temp+3	
	decsz	temp+4
	jmp	AddRangeLoop
	call	EndRange
	ret

SubRange
	call	PrepRange
	stc
SubRangeLoop
	mov	FSR,temp+2
	mov	temp+5,INDF
	mov	FSR,temp+3
	sub	INDF,temp+5
	inc	temp+2
	inc	temp+3	
	decsz	temp+4
	jmp	SubRangeLoop
	call	EndRange
	ret
SetRange
	call	PrepRange
SetRangeLoop
	mov	FSR,temp+2
	mov	temp+5,INDF
	mov	FSR,temp+3
	mov	INDF,temp+5
	inc	temp+2
	inc	temp+3	
	decsz	temp+4
	jmp	SetRangeLoop
	call	EndRange
	ret

GetRange
	call	PrepRange
GetRangeLoop
	mov	FSR,temp+3
	mov	temp+5,INDF
	mov	FSR,temp+2
	mov	INDF,temp+5
	inc	temp+2
	inc	temp+3	
	decsz	temp+4
	jmp	GetRangeLoop
	call	EndRange
	ret

cmd01	; readmem
	mov	W,#0
	jmp	getargs

arg01	; readmem
	mov	W,INDF	
	jz	nextcommand
	mov	FSR,INDF
	mov	temp,INDF
	mov	FSR,#30h
	mov	INDF,temp
	mov	W,#0
	jmp	getargs

cmd12
cmd11
cmd10	; Move
	mov	W,#11
	jmp	getargs

arg12	; Move and reset PID and trapeze dests
        mov     W,#_perr1
        call    ClrRange
        mov     W,#_perr2
        call    ClrRange
arg11   ; Move and reset trapeze dest
        mov     W,#_pos1L
        call    ClrRange
        mov     W,#_pos2L
        call    ClrRange
arg10	; Move
        mov     W,#_pos1
        call    AddRange
        mov     W,#_pos2
        call    AddRange
        mov     W,#_rspeed1
        call    SetRange
        mov     W,#_rspeed2
        call    SetRange
        bank    TRAP1_B
        mov     W,#0FCh
        and     flags,W
        bank    TRAP2_B
        and     flags,W
	jmp	nextcommand

cmd20	; SetAccel
	mov	W,#3
	jmp	getargs

arg20	; SetAccel
        mov     W,#_raccel1
        call    SetRange
        mov     W,#_raccel2
        call    SetRange
	jmp	nextcommand

cmdfd ; stopfast
        mov     W,#_ierr1
        call    ClrRange
        mov     W,#_ierr2
        call    ClrRange
        mov     W,#_speed1
        call    ClrRange
        mov     W,#_speed2
        call    ClrRange
cmdfe ; stopslow
        mov     W,#_rspeed1
        call    ClrRange
        mov     W,#_rspeed2
        call    ClrRange
        jmp     nextcommand

cmdff ; reset
        page    init
        jmp     init

	org 700h
_perr1	dw	(300h + PID1_B) | perr)
_perr2	dw	(300h + PID2_B) | perr)
_ierr1	dw	(400h + PID1_B) | ierr)
_ierr2	dw	(400h + PID2_B) | ierr)
_pos1L	dw	(400h + TRAP1_B) | (pos-1))
_pos2L	dw	(400h + TRAP2_B) | (pos-1))
_pos1	dw	(300h + TRAP1_B) | pos)
_pos2	dw	(300h + TRAP2_B) | pos)
_rspeed1 dw	(300h + TRAP1_B) | (rspeed-1))
_rspeed2 dw	(300h + TRAP2_B) | (rspeed-1))
_speed1	dw	(300h + TRAP1_B) | (speed-1))
_speed2	dw	(300h + TRAP2_B) | (speed-1))
_raccel1 dw	(200h + TRAP1_B) | (raccel-1))
_raccel2 dw	(200h + TRAP2_B) | (raccel-1))
_pwmspeed1 dw	(200h + INT_B) | speed1)
_pwmspeed2 dw	(200h + INT_B) | speed2)
_count1 dw	(200h + INT_B) | count1)
_count2 dw	(200h + INT_B) | count2)
_clock	dw	(200h + INT_B) | clock)




