; PID control for two wheel robot on SX (Ubicom) microcontroller.
; Copyright (C) 1999-2001 Blaise Gassend <blaise at gassend dot com>
; 
; This program is free software; you can redistribute it and/or
; modify it under the terms of the GNU General Public License
; as published by the Free Software Foundation; either version 2
; of the License, or (at your option) any later version.
; 
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
; 
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

; SPI as master!

; Constants
DOPID = 1
DOTRAPEZE = 1
DOEXTPID = 0

FRICTION = 10

IERRLIM = 000Fh
PCONSTH = 0
PCONST = 200
PCONSTL = 0
ICONSTH = 0
ICONST = 20
ICONSTL = 0
DCONSTH = 0
DCONST = 400 
DCONSTL = 0

PIDPERIOD = 080h

WAITINGFORCMD = 35h ; Do not change from 35h
SPIDELAYLENGTH = 0B0h
	device  sx18ac, pages4, banks8, oschs
	device  turbo, optionx, carryx

	reset   init

; Port A bits

PWM1	equ	0
DIR1	equ	1
PWM2	equ	2
DIR2	equ	3

; Port B bits

QA0     equ     0
QA1     equ     1
QB0     equ     2
QB1     equ     3
CK      equ     4
RX      equ     5
TX      equ     6
NSS	equ	7
MQA0    equ     1
MQA1    equ     2
MQB0    equ     4
MQB1    equ     8
MCK     equ     16
MRX     equ     32
MTX     equ     64
MNSS	equ	128

; Modes

;CMP     equ     08h
WKPND   equ     09h
WKED    equ     0Ah
WKEN    equ     0Bh
ST      equ     0Ch
LVL     equ     0Dh
PLP     equ     0Eh
DIR     equ     0Fh

; Banks

INT_B	equ	10h
SPI_B	equ	30h
PID1_B	equ	50h
PID2_B	equ	70h
TRAP1_B equ	90h
TRAP2_B	equ	0B0h

; SPI commands
CMDRESET equ	0FFh

; Ram

	org     7h
inttmp  ds      1
temp	ds	1

; Note : LSB first!

; bank 0
	org     10h
count1  ds      2
count2  ds      2
wkenech ds      1
wkedech ds      1
speed1 	ds      2
speed2  ds	2
pwmerr1 ds	2
pwmerr2 ds	2
clock	ds	2

; bank 1
	org	10h
iodata	ds	12
;	org	1Dh
spidelay ds	1
cmd	ds	1
iobyte	ds	1
iostat  ds	1  

; Banks 50h and 70h

org	9h ; locals

org	0dh
fspeed	ds	3 ; fracional part of speed
org	10h ; globals
ispeed  ds	2 ; integer part of speed
perr	ds	3
ierr	ds	4
derr	ds	2
dist	ds	2 ; moved since last time
abspos	ds	3 ; absolute position

; Banks 90h and 0B0h

org 8h
temp	ds	16
org 10h
	ds	1
pos	ds	3 ; 3:1
	ds	1
speed	ds	2 ; 2:1
	ds	1
rspeed	ds	2 ; 2:1
	ds	1
raccel	ds	1 ; 1:1
flags	ds	1
perrlow ds	1
temp2	ds	2

; iostat contents
; 0-3 curbyte
SPIRHDR	equ	4
SPIWAIT	equ	5
SPIDATA equ	6
SPICMD	equ	7
MSPIDATA equ	64

	include "macro.inc"
	include "pid.asm"
	include "trapeze.asm"
	include "sercmd.asm"

; Interrupt handler
	org     0
interrupt
	bank 0
; Handle PWM
	snb	RTCC.7
	jmp	PB1	; NotTime for PWM yet

	mov	W,#085h ; New clock pos...
	clc  
	add	W,RTCC
	mov	inttmp,W
	clc
	snb	ra.PWM1   ; Must have fixed number of ticks after RTCC read.
	add	pwmerr1,W ; Update pwm1err
	clr	W
	snb	ra.PWM1
	add	pwmerr1+1,W
	movb	ra.PWM1,pwmerr1+1.7
;	clc	; Shouldn't be necessary
	mov	W,inttmp	; Update pwm2err
	snb	ra.PWM2
	add	pwmerr2,W
	clr	W
	snb	ra.PWM2
	add	pwmerr2+1,W
	movb	ra.PWM2,pwmerr2+1.7

PWMH2	
	clc		; Add new time to clock
	mov	W,inttmp
	add	clock,W
	jnc	PWMH1
	inc	clock+1
	mov	W,#3
	and	W,clock+1
	jnz	PWMH1

	stc	
	sub	pwmerr1,speed1
	sub	pwmerr1+1,speed1+1
	stc
	sub	pwmerr2,speed2
	sub	pwmerr2+1,speed2+1

PWMH1 ; Handle SPI sending
	stc
	mov	W,inttmp
	sub 	RTCC,W

	bank  	#SPI_B	; Update SPI counter
	mov	W,/spidelay
	jz	PB1	; SPI is inactive
	clc
	mov	W,>>inttmp
	stc	
	sub	spidelay,W
	jc	PB1	; SPI counter not expired yet.
	call	SPISendBits

; Handle port b interrupts
PB1
	mov	FSR,M ; Auto bank 0
	mov     M,#WKPND
	clr     W
	mov     !rb,W   ; exchanges...
	mov     inttmp,W
	mov	M,FSR
	mov     W,/wkenech
	and     inttmp,W  ; ignore disabled bits
	mov	W,wkedech ; ### Only accept if change has remained.
	xor	W,rb ; ###
	or	W,#0Fh
	and	inttmp,W ; ###
	snz
	reti

; Quadrature decoder interrupt
	jnb     inttmp.QA0,QD1    ; QA0 changed
	mov     W,>>rb
	and     W,#MQA0
	xor     wkedech,W
	incdecb count1,wkedech.QA0
	setb    wkenech.QA0
	clrb    wkenech.QA1
	movb    wkedech.QA1,rb.QA1

QD1
	jnb     inttmp.QA1,QD2    ; QA1 changed
	mov     W,<<rb
	and     W,#MQA1
	xor     wkedech,W
	decincb count1,wkedech.QA1
	setb    wkenech.QA1
	clrb    wkenech.QA0
	movb    wkedech.QA0,rb.QA0

QD2
	jnb     inttmp.QB0,QD3    ; QB0 changed
	mov     W,>>rb
	and     W,#MQB0
	xor     wkedech,W
	incdecb count2,wkedech.QB0
	setb    wkenech.QB0
	clrb    wkenech.QB1
	movb    wkedech.QB1,rb.QB1

QD3
	jnb     inttmp.QB1,QD4    ; QB1 changed
	mov     W,<<rb
	and     W,#MQB1
	xor     wkedech,W
	decincb count2,wkedech.QB1
	setb    wkenech.QB1
	clrb    wkenech.QB0
	movb    wkedech.QB0,rb.QB0

QD4
;	jb	wkenech.NSS,intend
;	mov	W,wkedech
;	xor	W,rb
;	and	W,#MNSS
;	jnz	ssdetect ; We don't want to miss early changes to NSS!
	jnb	inttmp.NSS,intend ; Start transfer on NSS change
;ssdetect
	setb	wkenech.NSS ; Wait to send before enabling interrupt

	bank	#SPI_B
	or	iostat,#0Fh ; We are at the start of a byte

	mov	W,#0D0h ; if recieved command or parameters we wait.
	and	W,iostat
	xor	W,#0D0h
	sz	
	call	StartSPI

intend
; Update interrupt parameters
	mov	FSR,M ; Auto bank 0
        mov     M,#WKED
        mov     W,wkedech
        mov     !rb,W

        mov     M,#WKEN
        mov     W,wkenech
        mov     !rb,W
	mov	M,FSR

; End of interrupt
        reti

; ************************************************************************
; ************************************************************************
; ************************************************************************
; ************************************************************************

StartSPI
	clrb	rb.CK ; Send first bit

	mov	FSR,M ; Enable Tx and CK pin as output.
	mov	M,#DIR 
	mov	W,#10101111b
	mov	!rb,W
	mov	M,FSR

	bank	#SPI_B
	call	SPITx
	mov	spidelay,#SPIDELAYLENGTH
	ret

SPISendBits
	mov	W,#0Fh	;	Check if more to send.
	and	W,iostat
	jz	DoneSPI
	movb	rb.CK,iostat.0 ; Set new clock value
	dec	iostat 	; Read or Write to SPI and update clock. 	
	mov	spidelay,#SPIDELAYLENGTH
	jnb	iostat.0, SPIRx ; clock rise means read time

SPITx
	; write time, clock just rose
	movb	rb.TX,iobyte.7	; Output high bit
	ret

SPIRx
	; read time, clock just dropped
	movb	C,rb.RX	; Read
	rl	iobyte
	ret

DoneSPI
	; Copy byte if necessary
	
notarg
	mov	FSR,M ; Put Tx and CK pin as input.
	movb	wkedech.NSS,rb.NSS ; Update NSS status for end of transmit
	clrb	wkenech.NSS ; Prepare to detect next change
        mov     M,#WKED
        mov     W,wkedech
        mov     !rb,W
        mov     M,#WKEN
        mov     W,wkenech
        mov     !rb,W
	mov	M,FSR
	mov	M,#DIR 
	mov	W,#11111111b
	mov	!rb,W
	mov	M,FSR

	bank	#SPI_B
	mov	spidelay,#0FFh ; Signal that work is done.

	mov	FSR,#30h ; Cur byte in FSR
	mov	W,<>iostat 
	and	W,#0Fh
	add	FSR,W
	stc
	sub	iostat,#10h ; Update iostat

	mov	W,#0Ch ; Did I just read an arg?
	and	W,FSR
	xor	W,#0Ch
	snz	
	ret
	mov	INDF,iobyte ; Store arg

	dec	FSR ; Do I have more to reply?
	mov	W,#2Fh
	xor	W,FSR
	snz
	ret

	mov	iobyte,INDF
	ret	

; *************************************************************************
; *************************************************************************
; *************************************************************************
; *************************************************************************

init
; Clear memory
	clr	W
	mov	FSR,W
	mov	W,#10h
ClearMemoryLoop:
	or	FSR,W
	clr	INDF
	inc	FSR
	jnz	ClearMemoryLoop

; Set init
        bank    #0
        mov     W,#10010001b
        mov     !option,W

; Enable interrupts for quadrature decoding and serial.
        mov     M,#ST ; Schmitt trigger on port B
        clr     W
        mov     !rb,W

;	mov	M,#PLP ; Set pullups on rb
;	mov	W,#01110000b
;	mov	!rb,W

        mov     M,#WKPND ; Forget pending interrupts
        clr     W
        mov     !rb,W

        mov     M,#WKEN ; Enable interrupts for quadrature decoding and SPI
        mov     W,#01110000b
        mov     wkenech,W
        mov     !rb,W

        mov     M,#WKED ; Select interrupts on rise or fall as apropriate
        mov     W,rb
        mov     wkedech,W
        mov     !rb,W
 
	mov	M,#DIR
	mov	W,#11111111b ; Will have to set tx and ck to 1 when apropriate.
	mov	!rb,W

; Prepare serial
	bank	#SPI_B ; Set serial in command mode
	mov	iostat,#0E0h
	mov	spidelay,#0FFh
	mov	iobyte,#WAITINGFORCMD
IF DOEXTPID == 1
	clr	iostat
ENDIF

; Prepare PWM
	bank	#0
        mov     W,#DIR
        mov     M,W
        mov     !ra,#00000000b

; ********************************************************************
loop ; **************************** MAIN LOOP ************************
; ********************************************************************

IF DOEXTPID == 1
	bank	INT_B
StabilizeLoop
	mov	temp,count1
	mov	temp+1,count1+1
	mov	temp+2,count2
	mov	temp+3,count2+1
	mov	W,count1+1
	xor	W,temp+1
	jnz	StabilizeLoop
	mov	W,count2+1
	xor	W,temp+3
	jnz	StabilizeLoop
	bank	SPI_B
waitspidone
	jnb	rb.NSS,waitspidone
	mov	temp+4,iodata
	mov	temp+5,iodata+1
	mov	temp+6,iodata+2
	mov	temp+7,iodata+3
	mov	iodata,temp
	mov	iodata+1,temp+1
	mov	iodata+2,temp+2
	mov	iodata+3,temp+3
	test	iostat
	jnz	NoUpdate

	bank	INT_B
	sb	temp+5.7
	clrb	ra.DIR1
	jnb	temp+5.7,Speed1Pos
	not	temp+4
	not	temp+5
	stc
	clr	W
	add	temp+4,W
	add	temp+5,W
	setb	ra.DIR1
Speed1Pos
	mov	speed1,temp+4
	mov	speed1+1,temp+5
	sb	temp+7.7
	clrb	ra.DIR2
	jnb	temp+7.7,Speed2Pos
	not	temp+6
	not	temp+7
	stc
	clr	W
	add	temp+6,W
	add	temp+7,W
	setb	ra.DIR2
Speed2Pos
	mov	speed2,temp+6
	mov	speed2+1,temp+7
NoUpdate
	bank	SPI_B
	mov	W,#5h
	mov	iostat,W
ENDIF

IF DOTRAPEZE == 1
	page	CallTrapeze
	call	CallTrapeze
ENDIF
IF DOPID == 1
	page	CallPID
	call	CallPID
ENDIF

waitclock
IF DOEXTPID != 1
	page	ExecSerCmd
	call	ExecSerCmd
ENDIF
	bank	INT_B
	page	waitclock
	stc
	mov	W,#PIDPERIOD
	mov	W,(clock+1)-W
	jnc	waitclock
;	jnb	clock+1.7,waitclock ; Wait for 2ms to be up
;	jnb	clock+1.6,waitclock
;	mov	W,#0C0h 
	mov	W,#PIDPERIOD
	stc
	sub	clock+1,W

	page	loop
        jmp     loop

page0end

end

	

	


