; PID control for two wheel robot on SX (Ubicom) microcontroller.
; Copyright (C) 1999-2001 Blaise Gassend <blaise at gassend dot com>
; 
; This program is free software; you can redistribute it and/or
; modify it under the terms of the GNU General Public License
; as published by the Free Software Foundation; either version 2
; of the License, or (at your option) any later version.
; 
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
; 
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

; Constants
DOPID = 0
DOTRAPEZE = 0
DOEXTPID = 1

FRICTION = 128

IERRLIM = 01FFh
PCONSTH = 0
PCONST = 256 * 2
PCONSTL = 0
ICONSTH = 0
ICONST = 128
ICONSTL = 0
DCONSTH = 0
DCONST = 256 * 32 
DCONSTL = 0


	device  sx18ac, pages4, banks8, oschs
	device  turbo, optionx, carryx

	reset   init

; Port A bits

PWM1	equ	0
DIR1	equ	1
PWM2	equ	2
DIR2	equ	3

; Port B bits

QA0     equ     0
QA1     equ     1
QB0     equ     2
QB1     equ     3
CK      equ     4
RX      equ     5
TX      equ     6
NSS	equ	7
MQA0    equ     1
MQA1    equ     2
MQB0    equ     4
MQB1    equ     8
MCK     equ     16
MRX     equ     32
MTX     equ     64
MNSS	equ	128

; Modes

;CMP     equ     08h
WKPND   equ     09h
WKED    equ     0Ah
WKEN    equ     0Bh
ST      equ     0Ch
LVL     equ     0Dh
PLP     equ     0Eh
DIR     equ     0Fh

; Banks

INT_B	equ	10h
SPI_B	equ	30h
PID1_B	equ	50h
PID2_B	equ	70h
TRAP1_B equ	90h
TRAP2_B	equ	0B0h

; SPI commands
CMDRESET equ	0FFh

; Ram

	org     7h
inttmp  ds      1
temp	ds	1

; Note : LSB first!

; bank 0
	org     10h
count1  ds      2
count2  ds      2
wkenech ds      1
rbech 	ds      1
speed1 	ds      2
speed2  ds	2
pwmerr1 ds	2
pwmerr2 ds	2
clock	ds	2

; bank 1
	org	10h
iodata	ds	13
;	org	1Dh
cmd	ds	1
iobyte	ds	1
iostat  ds	1  

; iostat contents
; 0-3 curbyte
SPIRHDR	equ	4
SPIWAIT	equ	5
SPIDATA equ	6
SPICMD	equ	7
MSPIDATA equ	64

	include "macro.inc"
	include "pid.asm"
	include "trapeze.asm"
	include "sercmd.asm"

; Interrupt handler
	org     0
interrupt
	bank 0
; Handle PWM
	snb	RTCC.7
	jmp	PB1	; NotTime for PWM yet

	mov	W,#0A0h
	clc  
	add	W,RTCC
	mov	inttmp,W
	clc
	snb	ra.PWM1   ; Must have fixed number of ticks after RTCC read.
	add	pwmerr1,W ; Update pwm1err
	clr	W
	snb	ra.PWM1
	add	pwmerr1+1,W
	movb	ra.PWM1,pwmerr1+1.7
;	clc	; Shouldn't be necessary
	mov	W,inttmp	; Update pwm2err
	snb	ra.PWM2
	add	pwmerr2,W
	clr	W
	snb	ra.PWM2
	add	pwmerr2+1,W
	movb	ra.PWM2,pwmerr2+1.7

PWMH2	
	
	clc		; Add new time to clock
	mov	W,inttmp
	add	clock,W
	jnc	PWMH1
	inc	clock+1
	mov	W,#3
	and	W,clock+1
	jnz	PWMH1

	stc	
	sub	pwmerr1,speed1
	sub	pwmerr1+1,speed1+1
	stc
	sub	pwmerr2,speed2
	sub	pwmerr2+1,speed2+1

PWMH1
	clc
	mov	W,inttmp
	sub 	RTCC,W	

; Handle port b interrupts
PB1
	mov	rbech,rb
	mov	FSR,M
	mov     M,#WKPND
	clr     W
	mov     !rb,W   ; exchanges...
	mov     inttmp,W
	mov	M,FSR
	mov     W,/wkenech
	and     inttmp,W  ; ignore disabled bits
	mov	W,wkedech ; ### Only accept if change has remained.
	xor	W,rb ; ###
	or	W,#0Fh
	and	inttmp,W ; ###
	snb     Z
	reti

; Quadrature decoder interrupt
	jnb     inttmp.QA0,QD1    ; QA0 changed
	mov     W,>>rb
	and     W,#MQA0
	xor     wkedech,W
	incdecb count1,wkedech.QA0
	setb    wkenech.QA0
	clrb    wkenech.QA1
	movb    wkedech.QA1,rb.QA1

QD1
	jnb     inttmp.QA1,QD2    ; QA1 changed
	mov     W,<<rb
	and     W,#MQA1
	xor     wkedech,W
	decincb count1,wkedech.QA1
	setb    wkenech.QA1
	clrb    wkenech.QA0
	movb    wkedech.QA0,rb.QA0

QD2
	jnb     inttmp.QB0,QD3    ; QB0 changed
	mov     W,>>rb
	and     W,#MQB0
	xor     wkedech,W
	incdecb count2,wkedech.QB0
	setb    wkenech.QB0
	clrb    wkenech.QB1
	movb    wkedech.QB1,rb.QB1

QD3
	jnb     inttmp.QB1,QD4    ; QB1 changed
	mov     W,<<rb
	and     W,#MQB1
	xor     wkedech,W
	decincb count2,wkedech.QB1
	setb    wkenech.QB1
	clrb    wkenech.QB0
	movb    wkedech.QB0,rb.QB0

QD4
; Handle \SS pin
	jnb	inttmp.NSS,HSPI
	mov	W,#MNSS
	xor	wkedech,W
	jb	wkedech.NSS,SSUP

	; \SS just went low
	mov	FSR,M
	mov	M,#DIR ; Enable Tx pin as output.
	mov	W,#10111111b
	mov	!rb,W
	mov	M,FSR
	movb	wkedech.CK,rb.CK ; Prepare CK interrupts
	clrb	wkenech.CK

	; Choose byte to send.
IF DOEXTPID == 1
	mov	W,#SPI_B
	mov	FSR,W
	clc
	add	FSR,iostat
	dec	FSR
	mov	iobyte,INDF
	mov	W,iostat
	xor	W,#5
	jnz	NotInit
	mov	iobyte,#10101010b
NotInit
	test	iostat
	jnz	NotWait
	mov	iobyte,#01010101b
NotWait
ELSE
	mov	W,#30h
	mov	FSR,W
	mov	W,#0F0h
	and	W,iostat
	jz	WriteData
	mov	W,#10000000b
	snb	iostat.SPIDATA
	mov	W,#01111111b
	snb	iostat.SPIWAIT
	mov	W,#00000001b
	snb	iostat.SPIRHDR
	mov	W,#11111110b
	mov	iobyte,W
ENDIF
	jmp	SPITx

WriteData
	mov	W,iostat
	clc
	add	FSR,W
	mov	W,INDF
	mov	iobyte,W
	jmp	SPITx
	
SSUP ; \SS just went high
	mov	FSR,M
	mov	M,#DIR ; Disable Tx pin as output.
	mov	W,#11111111b
	mov	!rb,W
	mov	M,FSR
	setb	wkenech.CK

	; Handle input data.
IF DOEXTPID == 1
	mov	FSR,#SPI_B
	mov	W,iostat
	test	W
	jz	intend
	dec	iostat
	clc
	add	FSR,iostat
	mov	INDF,iobyte
	jmp	intend
ENDIF
	mov	W,#30h
	mov	FSR,W
	mov	W,iostat
	and	W,#0F0h
	jz	Replying
	jb	iostat.SPICMD,RecievedCommand
	jb	iostat.SPIDATA,RecieveData
	jb	iostat.SPIWAIT,WaitToReply
	; Sent header byte of reply
	clrb	iostat.SPIRHDR	

Replying
	test	iostat
	jz	ResetSPI
	dec	iostat ; Point to next byte
WaitToReply
	test	iobyte ; Check master sent 0
	jz	intend
ResetSPI
	clr	iostat ; Return to initial SPI mode
	setb	iostat.SPICMD
	jmp	intend
	
RecievedCommand
	mov	W,iobyte ; Check for 0h nop command
	jz	ResetSPI
	mov	W,/iobyte ; Check for 0FFh nop command
	jz	ResetSPI
	mov	W,iobyte ; Store command. 
	mov	cmd,W 
	and	W,#0Fh ; Command has parameters?
	xor	W,#0Ch
	jz	NormCommand ; 0Ch
	and	W,#0Ch
	jz	NoParam ; 0Dh to 0Fh
	
NormCommand
	mov	W,iobyte ; Prepare to read data bytes
	and	W,#0Fh
	or	W,#MSPIDATA
	mov	iostat,W
	jmp	intend

NoParam
	clr	iostat ; No more data to be read
	setb	iostat.SPIWAIT
	jmp	intend
	
RecieveData ; Recieved command parameter
	mov	W,iostat
	and	W,#0Fh	
	clc
	add	FSR,W
	mov	W,iobyte
	mov	INDF,W
	test	iostat
	jz	NoParam
	dec	iostat
	jmp	intend

HSPI
; Handle SPI IO
	jb	wkedech.NSS,intend ; Only SPI when \SS low
	jnb	inttmp.CK,intend
	mov	W,#MCK
	xor	wkedech,W ; Wait for next CK edge
	and	W,wkedech ; Check current state of CK edge
	bank	#30h
	jz	SPIRx ; clock drop means read time

SPITx
	; write time, clock just rose
	movb	rb.TX,iobyte.7	; Output high bit
	jmp	intend

SPIRx
	; read time, clock just dropped
	movb	C,rb.RX	; Read
	rl	iobyte

intend
; Update interrupt parameters
	bank	#0
	mov	FSR,M
        mov     M,#WKED
        mov     W,wkedech
        mov     !rb,W

        mov     M,#WKEN
        mov     W,wkenech
        mov     !rb,W
	mov	M,FSR

; End of interrupt
        reti

init
; Clear memory
	clr	W
	mov	FSR,W
	mov	W,#10h
ClearMemoryLoop:
	or	FSR,W
	clr	INDF
	inc	FSR
	jnz	ClearMemoryLoop

; Set init
        bank    #0
        mov     W,#10010000b
        mov     !option,W

; Enable interrupts for quadrature decoding and serial.
        mov     M,#ST ; Schmitt trigger on port B
        clr     W
        mov     !rb,W

	mov	M,#PLP ; Set pullups on rb
	mov	W,#01110000
	mov	!rb,W

        mov     M,#WKPND ; Forget pending interrupts
        clr     W
        mov     !rb,W

        mov     M,#WKEN ; Enable interrupts for quadrature decoding and SPI
        mov     W,#01110000b
        mov     wkenech,W
        mov     !rb,W

        mov     M,#WKED ; Select interrupts on rise or fall as apropriate
        mov     W,rb
        mov     wkedech,W
        mov     !rb,W
 
	mov	M,#DIR
	mov	W,#11111111b ; Will have to set tx to 1 when apropriate.
	mov	!rb,W

; Prepare serial
	bank	#SPI_B ; Set serial in command mode
	clr	iostat
	setb	iostat.SPICMD
IF DOEXTPID == 1
	clr	iostat
ENDIF

; Prepare PWM
	bank	#0
        mov     W,#DIR
        mov     M,W
        mov     !ra,#00000000b

loop

IF DOEXTPID == 1
	bank	INT_B
StabilizeLoop
	mov	temp,count1
	mov	temp+1,count1+1
	mov	temp+2,count2
	mov	temp+3,count2+1
	mov	W,count1+1
	xor	W,temp+1
	jnz	StabilizeLoop
	mov	W,count2+1
	xor	W,temp+3
	jnz	StabilizeLoop
	bank	SPI_B
waitspidone
	jnb	rb.NSS,waitspidone
	mov	temp+4,iodata
	mov	temp+5,iodata+1
	mov	temp+6,iodata+2
	mov	temp+7,iodata+3
	mov	iodata,temp
	mov	iodata+1,temp+1
	mov	iodata+2,temp+2
	mov	iodata+3,temp+3
	test	iostat
	jnz	NoUpdate

	bank	INT_B
	sb	temp+5.7
	clrb	ra.DIR1
	jnb	temp+5.7,Speed1Pos
	not	temp+4
	not	temp+5
	stc
	clr	W
	add	temp+4,W
	add	temp+5,W
	setb	ra.DIR1
Speed1Pos
	mov	speed1,temp+4
	mov	speed1+1,temp+5
	sb	temp+7.7
	clrb	ra.DIR2
	jnb	temp+7.7,Speed2Pos
	not	temp+6
	not	temp+7
	stc
	clr	W
	add	temp+6,W
	add	temp+7,W
	setb	ra.DIR2
Speed2Pos
	mov	speed2,temp+6
	mov	speed2+1,temp+7
NoUpdate
	bank	SPI_B
	mov	W,#5h
	mov	iostat,W
ENDIF

IF DOTRAPEZE == 1
	page	CallTrapeze
	call	CallTrapeze
ENDIF
IF DOPID == 1
	page	CallPID
	call	CallPID
ENDIF

waitclock
IF DOEXTPID != 1
	page	ExecSerCmd
	call	ExecSerCmd
ENDIF
	bank	INT_B
	page	waitclock
	jnb	clock+1.6,waitclock
	jnb	clock+1.5,waitclock
	mov	W,#60h
	stc
	sub	clock+1,W

	page	loop
        jmp     loop

page0end
end

	

	


