; PID control for two wheel robot on SX (Ubicom) microcontroller.
; Copyright (C) 1999-2001 Blaise Gassend <blaise at gassend dot com>
; 
; This program is free software; you can redistribute it and/or
; modify it under the terms of the GNU General Public License
; as published by the Free Software Foundation; either version 2
; of the License, or (at your option) any later version.
; 
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
; 
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

	device  pins28, pages4, banks8, ircdiv4
	device  turbo, optionx, carryx

	reset   init

	include "macro.inc"

; Port B bits

QA0     equ     0
QA1     equ     1
QB0     equ     2
QB1     equ     3
CK      equ     4
RX      equ     5
TX      equ     6
MQA0    equ     1
MQA1    equ     2
MQB0    equ     4
MQB1    equ     8
MCK     equ     16
MRX     equ     32
MTX     equ     64

; Modes

CMP     equ     08h
WKPND   equ     09h
WKED    equ     0Ah
WKEN    equ     0Bh
ST      equ     0Ch
LVL     equ     0Dh
PLP     equ     0Eh
DIR     equ     0Fh

; Ram

	org     8h
temp    ds      1
temp2	ds	1

; Note : LSB first!

; bank 0
	org     10h
count1  ds      2
count2  ds      2
wkenech ds      1
wkedech ds      1
speed1 	ds      2
speed1  ds	2
outerr1 ds	2
outerr2 ds	2
clock	ds	2

; bank 1
	org	10h

; Interrupt handler
	org     0
interrupt
	bank #0
; Handle PWM
	snb	RTCC,7
	jmp	PB1	; NotTime for PWM yet

	mov	w,/RTCC
	
	
			

; Handle port b interrupts
PB1
	mov     M,#WKPND
	clr     W
	mov     !rb,W   ; exchanges...
	mov     temp,W
	mov     W,/wkenech
	and     temp,W  ; ignore disabled bits
	snb     Z
	reti

; Quadrature decoder interrupt
	jnb     temp.QA0,QD1    ; QA0 changed
	mov     W,>>rb
	and     W,#MQA0
	xor     wkedech,W
	incdecb count1,wkedech.QA0
	setb    wkenech.QA0
	clrb    wkenech.QA1
	movb    wkedech.QA1,rb.QA1

QD1
	jnb     temp.QA1,QD2    ; QA1 changed
	mov     W,<<rb
	and     W,#MQA1
	xor     wkedech,W
	decincb count1,wkedech.QA1
	setb    wkenech.QA1
	clrb    wkenech.QA0
	movb    wkedech.QA0,rb.QA0

QD2
	jnb     temp.QB0,QD3    ; QB0 changed
	mov     W,>>rb
	and     W,#MQB0
	xor     wkedech,W
	incdecb count2,wkedech.QB0
	setb    wkenech.QB0
	clrb    wkenech.QB1
	movb    wkedech.QB1,rb.QB1

QD3
	jnb     temp.QB1,QD4    ; QB1 changed
	mov     W,<<rb
	and     W,#MQB1
	xor     wkedech,W
	decincb count2,wkedech.QB1
	setb    wkenech.QB1
	clrb    wkenech.QB0
	movb    wkedech.QB0,rb.QB0

; Update interrupt parameters
        mov     M,#WKED
        mov     W,wkedech
        mov     !rb,W

        mov     M,#WKEN
        mov     W,#0F5h
        mov     W,wkenech
        mov     !rb,W

; End of interrupt
intend
        reti


init
        bank    #0
; Enable interrupts for quadrature decoding.
        mov     M,#WKPND
        clr     W
        mov     !rb,W

        mov     M,#WKED
        mov     W,rb
        mov     wkedech,W
        mov     !rb,W
 
        mov     M,#WKEN
        mov     W,#0F5h
        mov     wkenech,W
        mov     !rb,W

        mov     W,#10000111b
        mov     !option,W
        mov     W,#0Fh
        mov     M,W
        mov     !ra,#00000000b
        clr     speed
        mov     W,#99h
        mov     ra,W
loop    mov     W,#2
        add     ra,W
        jmp     loop
