; PID control for two wheel robot on SX (Ubicom) microcontroller.
; Copyright (C) 1999-2001 Blaise Gassend <blaise at gassend dot com>
; 
; This program is free software; you can redistribute it and/or
; modify it under the terms of the GNU General Public License
; as published by the Free Software Foundation; either version 2
; of the License, or (at your option) any later version.
; 
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
; 
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

	device  sx18ac, pages4, banks8, oschs
	device  turbo, optionx, carryx

	reset   init

	include "macro.inc"
	include "pid.asm"

; Port A bits

PWM1	equ	0
DIR1	equ	1
PWM2	equ	2
DIR2	equ	3

; Port B bits

QA0     equ     0
QA1     equ     1
QB0     equ     2
QB1     equ     3
CK      equ     4
RX      equ     5
TX      equ     6
NSS	equ	7
MQA0    equ     1
MQA1    equ     2
MQB0    equ     4
MQB1    equ     8
MCK     equ     16
MRX     equ     32
MTX     equ     64
MNSS	equ	128

; Modes

;CMP     equ     08h
WKPND   equ     09h
WKED    equ     0Ah
WKEN    equ     0Bh
ST      equ     0Ch
LVL     equ     0Dh
PLP     equ     0Eh
DIR     equ     0Fh

; SPI commands
CMDRESET equ	0FFh

; Ram

	org     7h
inttmp  ds      1
temp	ds	1

; Note : LSB first!

; bank 0
	org     10h
count1  ds      2
count2  ds      2
wkenech ds      1
wkedech ds      1
speed1 	ds      2
speed2  ds	2
pwmerr1 ds	2
pwmerr2 ds	2
clock	ds	2

; bank 1
	org	10h
iodata	ds	13
;	org	1Dh
cmd	ds	1
iobyte	ds	1
iostat  ds	1  

; iostat contents
; 0-3 curbyte
SPIRHDR	equ	4
SPIWAIT	equ	5
SPIDATA equ	6
SPICMD	equ	7
MSPIDATA equ	64

; Interrupt handler
	org     0
interrupt
	bank 0
; Handle PWM
	snb	RTCC.7
	jmp	PB1	; NotTime for PWM yet

	mov	W,#0A0h
	clc  
	add	W,RTCC
	mov	inttmp,W
	clc
	snb	ra.PWM1   ; Must have fixed number of ticks after RTCC read.
	add	pwmerr1,W ; Update pwm1err
	clr	W
	snb	ra.PWM1
	add	pwmerr1+1,W
;	clc	; Shouldn't be necessary
	mov	W,inttmp	; Update pwm2err
	snb	ra.PWM2
	add	pwmerr2,W
	clr	W
	snb	ra.PWM2
	add	pwmerr2+1,W
	movb	ra.PWM1,pwmerr1+1.7
	movb	ra.PWM2,pwmerr2+1.7

PWMH2	
	
	clc		; Add new time to clock
	mov	W,inttmp
	add	clock,W
	jnc	PWMH1
	inc	clock+1
	mov	W,#3
	and	W,clock+1
	jnz	PWMH1

	stc	
	sub	pwmerr1,speed1
	sub	pwmerr1+1,speed1+1
	stc
	sub	pwmerr2,speed2
	sub	pwmerr2+1,speed2+1

PWMH1
	clc
	mov	W,inttmp
	sub 	RTCC,W	

; Handle port b interrupts
PB1
	mov     M,#WKPND
	clr     W
	mov     !rb,W   ; exchanges...
	mov     inttmp,W
	mov     W,/wkenech
	and     inttmp,W  ; ignore disabled bits
	mov	W,wkedech ; ### Only accept if change has remained.
	xor	W,rb ; ###
	or	W,#0Fh
	and	inttmp,W ; ###
	snb     Z
	reti

; Quadrature decoder interrupt
	jnb     inttmp.QA0,QD1    ; QA0 changed
	mov     W,>>rb
	and     W,#MQA0
	xor     wkedech,W
	incdecb count1,wkedech.QA0
	setb    wkenech.QA0
	clrb    wkenech.QA1
	movb    wkedech.QA1,rb.QA1

QD1
	jnb     inttmp.QA1,QD2    ; QA1 changed
	mov     W,<<rb
	and     W,#MQA1
	xor     wkedech,W
	decincb count1,wkedech.QA1
	setb    wkenech.QA1
	clrb    wkenech.QA0
	movb    wkedech.QA0,rb.QA0

QD2
	jnb     inttmp.QB0,QD3    ; QB0 changed
	mov     W,>>rb
	and     W,#MQB0
	xor     wkedech,W
	incdecb count2,wkedech.QB0
	setb    wkenech.QB0
	clrb    wkenech.QB1
	movb    wkedech.QB1,rb.QB1

QD3
	jnb     inttmp.QB1,QD4    ; QB1 changed
	mov     W,<<rb
	and     W,#MQB1
	xor     wkedech,W
	decincb count2,wkedech.QB1
	setb    wkenech.QB1
	clrb    wkenech.QB0
	movb    wkedech.QB0,rb.QB0

QD4
; Handle \SS pin
	jnb	inttmp.NSS,HSPI
	mov	W,#MNSS
	xor	wkedech,W
	jb	wkedech.NSS,SSUP

	; \SS just went low
	mov	M,#DIR ; Enable Tx pin as output.
	mov	W,#10111111b
	mov	!rb,W
	movb	wkedech.CK,rb.CK ; Prepare CK interrupts
	clrb	wkenech.CK

	; Choose byte to send.
	mov	W,#30h
	mov	FSR,W
	mov	W,#0F0h
	and	W,iostat
	jz	WriteData
	mov	W,#10000000b
	snb	iostat.SPIDATA
	mov	W,#01111111b
	snb	iostat.SPIWAIT
	mov	W,#00000001b
	snb	iostat.SPIRHDR
	mov	W,#11111110b
	mov	iobyte,W
	jmp	SPITx

WriteData
	mov	W,iostat
	clc
	add	FSR,W
	mov	W,INDF
	mov	iobyte,W
	jmp	SPITx
	
SSUP ; \SS just went high
	mov	M,#DIR ; Disable Tx pin as output.
	mov	W,#11111111b
	mov	!rb,W
	setb	wkenech.CK

	; Handle input data.
	mov	W,#30h
	mov	FSR,W
	mov	W,iostat
	and	W,#0F0h
	jz	Replying
	jb	iostat.SPICMD,RecievedCommand
	jb	iostat.SPIDATA,RecieveData
	jb	iostat.SPIWAIT,WaitToReply
	; Sent header byte of reply
	clrb	iostat.SPIRHDR	

Replying
	test	iostat
	jz	ResetSPI
	dec	iostat ; Point to next byte
WaitToReply
	test	iobyte ; Check master sent 0
	jz	intend
ResetSPI
	clr	iostat ; Return to initial SPI mode
	setb	iostat.SPICMD
	jmp	intend
	
RecievedCommand
	mov	W,iobyte ; Check for 0h nop command
	jz	ResetSPI
	mov	W,/iobyte ; Check for 0FFh nop command
	jz	ResetSPI
	mov	W,iobyte ; Store command. 
	mov	cmd,W 
	and	W,#0Fh ; Command has parameters?
	xor	W,#0Ch
	jz	NormCommand ; 0Ch
	and	W,#0Ch
	jz	NoParam ; 0Dh to 0Fh
	
NormCommand
	mov	W,iobyte ; Prepare to read data bytes
	and	W,#0Fh
	or	W,#MSPIDATA
	mov	iostat,W
	jmp	intend

NoParam
	clr	iostat ; No more data to be read
	setb	iostat.SPIWAIT
	jmp	intend
	
RecieveData ; Recieved command parameter
	mov	W,iostat
	and	W,#0Fh	
	clc
	add	FSR,W
	mov	W,iobyte
	mov	INDF,W
	test	iostat
	jz	NoParam
	dec	iostat
	jmp	intend

HSPI
; Handle SPI IO
	jb	wkedech.NSS,intend ; Only SPI when \SS low
	jnb	inttmp.CK,intend
	mov	W,#MCK
	xor	wkedech,W ; Wait for next CK edge
	and	W,wkedech ; Check current state of CK edge
	bank	#30h
	jz	SPIRx ; clock drop means read time

SPITx
	; write time, clock just rose
	movb	rb.TX,iobyte.7	; Output high bit
	jmp	intend

SPIRx
	; read time, clock just dropped
	movb	C,rb.RX	; Read
	rl	iobyte

intend
; Update interrupt parameters
	bank	#0
        mov     M,#WKED
        mov     W,wkedech
        mov     !rb,W

        mov     M,#WKEN
        mov     W,wkenech
        mov     !rb,W

; End of interrupt
        reti

init
        bank    #0
        mov     W,#10010000b
        mov     !option,W

; Enable interrupts for quadrature decoding and serial.
        mov     M,#ST
        clr     W
        mov     !rb,W

        mov     M,#WKPND
        clr     W
        mov     !rb,W

        mov     M,#WKED
        mov     W,rb
        mov     wkedech,W
        mov     !rb,W
 
        mov     M,#WKEN
        mov     W,#01110101b
        mov     wkenech,W
        mov     !rb,W

	mov	M,#DIR
	mov	W,#11111111b ; Will have to set tx to 1 when apropriate.
	mov	!rb,W

	bank	#30h
	clr	iostat
	setb	iostat.SPICMD
	bank	#0

; Prepare PWM
        mov     W,#DIR
        mov     M,W
        mov     !ra,#00000000b

	clr	count1
	clr	count2
        clr     speed1
        clr     speed1+1
        clr     speed2
        clr     speed2+1
	clr	pwmerr1
	clr	pwmerr1+1
	clr	pwmerr2
	clr	pwmerr2+1
	clr	clock
	clr	clock+1
        clr     ra

	;TMP
	mov	speed1+1,#1
	mov	speed2+1,#2
	;TMP

; Prepare PID
	bank #50h
	clr	fspeed
	clr	fspeed+1
	clr	fspeed+2
	clr	ispeed
	clr	ispeed+1
	clr	perr
	clr	perr+1
	clr	perr+2
	clr	ierr
	clr	ierr+1
	clr	ierr+2
	clr	ierr+3
	clr	derr
	clr	derr+1
	clr	dist
	clr	dist+1

loop
	bank	#0
	mov	W,count1
	mov	speed2,W
	mov	W,count1+1
	and	W,#3
	mov	speed2+1,W

	bank	#30h
	jnb	iostat.SPIWAIT,noin
	mov	W,iodata
	mov	temp,W
	mov	W,iodata+1
	mov	temp+1,W
	bank	#0h
	clc
	mov	W,temp
	mov	speed1,W
	mov	W,temp+1
	and	W,#3
	mov	speed1+1,W
	mov	temp,count1
	mov	temp+1,count1+1
	bank	#30h
	mov	iodata,temp
	mov	iodata+1,temp+1
	mov	W,#12h
	mov	iostat,W
noin
	
;	bank	#50h
;	page	#200h
;        mov     W,#11011111b
;        mov     !option,W
;	call	PIDFilter

	page	#0
        jmp     loop

end

