; PID control for two wheel robot on SX (Ubicom) microcontroller.
; Copyright (C) 1999-2001 Blaise Gassend <blaise at gassend dot com>
; 
; This program is free software; you can redistribute it and/or
; modify it under the terms of the GNU General Public License
; as published by the Free Software Foundation; either version 2
; of the License, or (at your option) any later version.
; 
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
; 
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

; flag bits
noovershoot equ	0 	; set to one we we leave overshoot condition to 
			; ensure that it won't overshoot again
;Removed, has problems
;nostop	equ	1	; we just want to goooooo...

; (velocity, position) space is divided into regions of +accel and -accel.
; They are:
;  Overshoot -> reduce speed.
;  Too fast -> reduce speed, don't go under limit speed.
;  Too slow -> accelerate don't go over limit speed or cross position = 0
; There are two zones of each type, which exchange when velocity and
; position change signs.
;
; To determine zone use following tests:
;  1) Velocity sign.
;  2) Check overshoot (approx : -2*a*d > v^2)
;			(d is negative)
;  if not overshoot:
;  3) Above or below maximum speed?

org 400h

CallTrapeze
	bank	TRAP1_B ; Call trapeze generation on line 1
	page	GenerateTrapeze
	call	GenerateTrapeze
	bank	TRAP2_B ; Call trapeze generation on line 2
	page	GenerateTrapeze
	call	GenerateTrapeze
	mov	temp,speed ; Store speeds in temporary
	mov	temp+1,speed+1
	clc	; Keep track of fractionnal errors in position.
	add	perrlow,speed-1 
	clr	W
	add	temp,W
	add	temp+1,W
	bank	TRAP1_B
	mov	temp+2,speed
	mov	temp+3,speed+1
	clc	; Keep track of fractionnal errors in position.
	add	perrlow,speed-1 
	clr	W
	add	temp+2,W
	add	temp+3,W
	bank	PID1_B ; Update PID1
	stc
	sub	perr,temp+2
	sub	perr+1,temp+3
	clr	W
	snb	temp+3.7
	mov	W,#0FFh
	sub	perr+2,W
	bank	PID2_B ; Update PID2
	stc
	sub	perr,temp
	sub	perr+1,temp+1
	clr	W
	snb	temp+1.7
	mov	W,#0FFh
	sub	perr+2,W
	ret

FlipSpeedAndPos
	not	speed-1
	not	speed
	not	speed+1	
	clr	W
	stc
	add	speed-1,w
	add	speed,w
	add	speed+1,w
	not	pos-1
	not	pos
	not	pos+1	
	not	pos+2
	clr	w
	stc
	add	pos-1,w
	add	pos,w
	add	pos+1,w
	add	pos+2,w
	ret

GenerateTrapeze
	jnb	speed+1.7,PositiveVelocity ; Test 1)
	; We'll just flip speed and pos, calculate, then re-flip...
	call	FlipSpeedAndPos
	call	GenerateTrapeze
	call	FlipSpeedAndPos
	ret

PositiveVelocity

;	jb	flags.nostop,KeepGoing ; If unlimited run, this is easy...
	; 2) Check for overshoot
	; if position is positive, then overshoot
	jb	pos+2.7,PosNegative
	mov	W,pos-1
	or	W,pos
	or	W,pos+1
	or	W,pos+2
	jnz	OverShot
	or	W,speed-1
	or	W,speed
	or	W,speed+1
	jnz	OverShot
	ret
PosNegative
	; Calculate -pos * raccel
	clr	temp
	clr	temp+1
	clr	temp+2
	clr	temp+3
	clr	temp+4
	clr	temp+5
	mov	W,#16
	mov	temp+6,W
MultLoop
	mov	W,<<raccel
	rl	raccel-1
	rl	raccel
	jnc	MultNoBit
	stc
	sub	temp,pos-1
	sub	temp+1,pos
	sub	temp+2,pos+1
	sub	temp+3,pos+2
	mov	W,#0FFh
	sub	temp+4,W
	sub	temp+5,W
MultNoBit
	dec	temp+6		
	snz
	jmp	MultDone
	clc
	rl	temp
	rl	temp+1
	rl	temp+2
	rl	temp+3
	rl	temp+4
	rl	temp+5
	jmp	MultLoop
MultDone
	; Now subtract v^2/2
	; I am detecting four values as equal. Oh well...
	mov	W,#24
	mov	temp+6,W
	mov	temp2,speed-1
	mov	temp2+1,speed
	mov	temp+7,speed+1
	clc			; Add 2*accel
	add	temp2,raccel-1
	add	temp2+1,raccel
	clr	W
	add	temp+7,W
	clc
	add	temp2,raccel-1
	add	temp2+1,raccel
	clr	W
	add	temp+7,W
	clc ; positive sign for now
SquareSub
	rr	temp+5
	rr	temp+4
	rr	temp+3
	rr	temp+2
	rr	temp+1
	rr	temp
	clc
	rr	temp+7
	rr	temp2+1
	rr	temp2
	jnc	SquareSubNoBit
	stc
	sub	temp,temp2
	sub	temp+1,temp2+1
	sub	temp+2,temp+7
	clr	W
	sub	temp+3,W
	sub	temp+4,W
	sub	temp+5,W
SquareSubNoBit
	dec	temp+6
	snz
	jmp	SquareSubDone
	jnb	temp2.0,SquareSubShift
	clr	W
	clc
	sub	temp,W
	sub	temp+1,W
	sub	temp+2,W
	sub	temp+3,W
	sub	temp+4,W
	sub	temp+5,W
SquareSubShift
	mov	W,<<temp+5 ; preserve sign when shifting
	rr	temp+5
	rr	temp+4
	rr	temp+3
	rr	temp+2
	rr	temp+1
	rr	temp
	mov	W,<<temp+5 ; preserve sign when shifting
	jmp	SquareSub

; SubDone is lower down...

TstSpdTooHigh ; nc -> lower or equal to max speed
	stc ; Check if velocity not too high
	mov	W,rspeed-1
	mov	W,speed-1-W
	mov	W,rspeed
	mov	W,speed-W
	mov	W,rspeed+1
	mov	W,speed+1-W
	ret

AddAccel
	clc
	mov	W,raccel-1
	add	speed-1,raccel-1
	mov	W,raccel
	add	speed,raccel
	clr	W
	add	speed+1,W
	ret
	
SetToMaxSpeed
	mov	speed-1,rspeed-1
	mov	speed,rspeed
	mov	speed+1,rspeed+1
	ret

UpdatePos
	clc
	add	pos-1,speed-1
	add	pos,speed
	add	pos+1,speed+1
	clr	W
	snb	speed+1.7
	mov	W,#0FFh
	add	pos+2,W
; Removed, has bugs
;;	snb	flags.nostop 
;;	ret
	sb	flags.noovershoot
	ret
	snb	pos+2.7
	ret
	stc
	sub	speed-1,pos-1
	sub	speed,pos
	sub	speed+1,pos+1
	clr	pos-1
	clr	pos
	clr	pos+1
	ret

SubAccel
	stc
	mov	W,raccel-1
	sub	speed-1,raccel-1
	mov	W,raccel
	sub	speed,raccel
	clr	W
	sub	speed+1,W
	sb	speed+1.1
	ret
	; If change of direction at very low speed.
	clc
	mov	W,speed-1	
	add	W,rspeed-1
	mov	W,speed
	add	W,rspeed
	mov	W,speed+1	
	add	W,rspeed+1
	snc
	ret
	clr	speed-1
	clr	speed
	clr	speed+1
	stc
	sub	speed-1,rspeed-1
	sub	speed,rspeed
	sub	speed+1,rspeed+1
	ret

SquareSubDone
	jb	temp+5.7,Overshot
	setb	flags.noovershoot
KeepGoing
	call 	TstSpdTooHigh
	jc	SpeedTooHigh
	call 	AddAccel
	call 	TstSpdTooHigh
	snc
	call	SetToMaxSpeed
	call	UpdatePos
	ret	
	
SpeedTooHigh
	call	SubAccel
	call	TstSpdTooHigh
	sc
	call	SetToMaxSpeed
	call	UpdatePos
	ret

OverShot
	call 	SubAccel
	call	UpdatePos
	ret
	

pagetrapezeend
